// Copyright (c) 2021 Terminus, Inc.
//
// This program is free software: you can use, redistribute, and/or modify
// it under the terms of the GNU Affero General Public License, version 3
// or later ("AGPL"), as published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

package service

import (
	"encoding/json"
	"strings"
	"time"

	"github.com/pkg/errors"
	log "github.com/sirupsen/logrus"

	"github.com/erda-project/erda/modules/hepa/common/util"
	. "github.com/erda-project/erda/modules/hepa/common/vars"
	"github.com/erda-project/erda/modules/hepa/repository/orm"
	"github.com/erda-project/erda/pkg/discover"
)

type AdminProjectDto struct {
	Name          string            `json:"name"`
	ClusterConfig map[string]string `json:"clusterConfig"`
}

type AdminRespDto struct {
	Success bool            `json:"success"`
	Data    AdminProjectDto `json:"data"`
}

const (
	CT_K8S  = "kubernetes"
	CT_DCOS = "dcos"
	CT_EDAS = "edas"
)

type ClusterRespDto struct {
	Success bool           `json:"success"`
	Data    ClusterInfoDto `json:"data"`
}

type ClusterInfoDto struct {
	DiceClusterType string `json:"DICE_CLUSTER_TYPE"`
	DiceRootDomain  string `json:"DICE_ROOT_DOMAIN"`
	MasterAddr      string `json:"MASTER_VIP_ADDR"`
	NetportalUrl    string `json:"NETPORTAL_URL"`
}

type GatewayAzInfoServiceImpl struct {
	engine *orm.OrmEngine
}

func NewGatewayAzInfoServiceImpl() (*GatewayAzInfoServiceImpl, error) {
	engine, error := orm.GetSingleton()
	if error != nil {
		return nil, errors.Wrap(error, "new GatewayAzInfoServiceImpl failed")
	}
	return &GatewayAzInfoServiceImpl{engine}, nil
}

func (impl *GatewayAzInfoServiceImpl) GetAz(cond *orm.GatewayAzInfo) (string, error) {
	azInfo, err := impl.GetAzInfo(cond)
	if err != nil {
		return "", err
	}
	return azInfo.Az, nil
}

func (impl *GatewayAzInfoServiceImpl) SelectByAny(cond *orm.GatewayAzInfo) ([]orm.GatewayAzInfo, error) {
	var result []orm.GatewayAzInfo
	if cond == nil {
		return result, errors.New(ERR_INVALID_ARG)
	}
	err := orm.SelectByAny(impl.engine, &result, cond)
	if err != nil {
		return result, errors.Wrap(err, ERR_SQL_FAIL)
	}
	return result, nil
}

func (impl *GatewayAzInfoServiceImpl) SelectValidAz() ([]orm.GatewayAzInfo, error) {
	var result []orm.GatewayAzInfo
	err := orm.Select(impl.engine.Distinct("az"), &result, `master_addr != ""`)
	if err != nil {
		return result, errors.Wrap(err, ERR_SQL_FAIL)
	}
	return result, nil
}

func fillInfo(info *orm.GatewayAzInfo, clusterInfo ClusterInfoDto) {
	clusterType := clusterInfo.DiceClusterType
	switch clusterType {
	case CT_K8S:
		info.Type = orm.AT_K8S
	case CT_DCOS:
		info.Type = orm.AT_DCOS
	case CT_EDAS:
		info.Type = orm.AT_EDAS
	default:
		info.Type = orm.AT_UNKNOWN
	}
	info.WildcardDomain = clusterInfo.DiceRootDomain
	if clusterInfo.MasterAddr == "" {
		info.MasterAddr = ""
	}
	if clusterInfo.NetportalUrl == "" {
		info.MasterAddr = clusterInfo.MasterAddr
	} else {
		info.MasterAddr = clusterInfo.NetportalUrl + "/" + clusterInfo.MasterAddr
	}
}

func (impl *GatewayAzInfoServiceImpl) GetAzInfoByClusterName(name string) (*orm.GatewayAzInfo, error) {
	info := &orm.GatewayAzInfo{
		Az: name,
	}
	code, body, err := util.CommonRequest("GET", discover.Scheduler()+"/api/clusterinfo/"+name, nil, map[string]string{"Internal-Client": "hepa-gateway"})
	if code >= 300 {
		err = errors.Errorf("get cluster info failed, code:%d", code)
	}
	if err != nil {
		return nil, err
	}
	clusterResp := &ClusterRespDto{}
	err = json.Unmarshal(body, clusterResp)
	if err != nil {
		return nil, errors.Wrapf(err, "unmarshal failed:%s", body)
	}
	if !clusterResp.Success {
		return nil, errors.Errorf("request cluster info failed: resp[%s]", body)
	}
	fillInfo(info, clusterResp.Data)
	return info, nil
}

func (impl *GatewayAzInfoServiceImpl) GetAzInfo(cond *orm.GatewayAzInfo) (*orm.GatewayAzInfo, error) {
	if cond == nil || cond.ProjectId == "" || cond.Env == "" {
		return nil, errors.New(ERR_INVALID_ARG)
	}
	info := &orm.GatewayAzInfo{}
	exist, err := orm.GetByAny(impl.engine, info, cond)
	if err != nil {
		return nil, errors.WithStack(err)
	}
	now := time.Now()
	if exist && info.Type != "" && (info.NeedUpdate == 0 || (now.Sub(info.UpdateTime).Seconds() < 60 &&
		now.Sub(info.UpdateTime).Seconds() > 0)) {
		return info, nil
	}
	code, body, err := util.CommonRequest("GET", discover.CoreServices()+"/api/projects/"+cond.ProjectId, nil,
		map[string]string{"Internal-Client": "hepa-gateway"})
	if err != nil {
		err = errors.WithMessage(err, "request dice admin failed")
		goto failback
	}
	if code < 300 {
		data := &AdminRespDto{}
		err = json.Unmarshal(body, data)
		if err != nil {
			err = errors.Wrapf(err, "unmarshal failed:%s", body)
			goto failback
		}
		if !data.Success {
			err = errors.Errorf("request dice admin failed: resp[%s]", body)
			goto failback
		}
		az, ok := data.Data.ClusterConfig[strings.ToUpper(cond.Env)]
		if !ok {
			err = errors.Errorf("can't find az of info[%+v] in admin resp[%s]", cond, body)
			goto failback
		}
		code, body, err = util.CommonRequest("GET", discover.Scheduler()+"/api/clusterinfo/"+az, nil, map[string]string{"Internal-Client": "hepa-gateway"})
		if code >= 300 || err != nil {
			goto failback
		}
		clusterResp := &ClusterRespDto{}
		err = json.Unmarshal(body, clusterResp)
		if err != nil {
			err = errors.Wrapf(err, "unmarshal failed:%s", body)
			goto failback
		}
		if !clusterResp.Success {
			err = errors.Errorf("request cluster info failed: resp[%s]", body)
			goto failback
		}
		fillInfo(info, clusterResp.Data)
		info.Az = az
		if exist {
			_, _ = orm.Update(impl.engine, info, "az", "wildcard_domain", "type", "master_addr")
		} else {
			cond.NeedUpdate = 1
			cond.Az = az
			_, _ = orm.Insert(impl.engine, cond)
		}
		return info, nil
	}

failback:
	if exist {
		log.Error(err)
		return info, nil
	}
	return nil, err
}
